@@ -235,7 +235,7 @@ md5hash_prev_conffile(struct pkginfo *pkg, char *oldhash, const char *oldname,
                              &otherpkg->configversion) != 0)
       continue;
     for (conff = otherpkg->installed.conffiles; conff; conff = conff->next) {
-      if (conff->obsolete)
+      if (conff->obsolete || conff->remove_on_upgrade)
         continue;
       if (strcmp(conff->name, namenode->name) == 0)
         break;
@@ -667,7 +667,7 @@ tarobject(struct tar_archive *tar, struct tar_entry *ti)
 {
   static struct varbuf conffderefn, symlinkfn;
   const char *usename;
-  struct fsys_namenode *usenode;
+  struct fsys_namenode *namenode, *usenode;
 
   struct conffile *conff;
   struct tarcontext *tc = tar->ctx;
@@ -688,12 +688,17 @@ tarobject(struct tar_archive *tar, struct tar_entry *ti)
   if (strchr(ti->name, '\n'))
     ohshit(_("newline not allowed in archive object name '%.255s'"), ti->name);
 
+  namenode = fsys_hash_find_node(ti->name, 0);
+
+  if (namenode->flags & FNNF_RM_CONFF_ON_UPGRADE)
+    ohshit(_("conffile '%s' marked for removal on upgrade, shipped in package"),
+           ti->name);
+
   /* Append to list of files.
    * The trailing ‘/’ put on the end of names in tarfiles has already
    * been stripped by tar_extractor(). */
   oldnifd = tc->newfiles_queue->tail;
-  nifd = tar_fsys_namenode_queue_push(tc->newfiles_queue,
-                                     fsys_hash_find_node(ti->name, 0));
+  nifd = tar_fsys_namenode_queue_push(tc->newfiles_queue, namenode);
   nifd->namenode->flags |= FNNF_NEW_INARCHIVE;
 
   debug(dbg_eachfile,
@@ -1246,6 +1251,19 @@ try_deconfigure_can(bool (*force_p)(struct deppossi *), struct pkginfo *pkg,
         return 0;
       }
     }
+    if (pkg->installed.is_protected) {
+      if (in_force(FORCE_REMOVE_PROTECTED)) {
+        warning(_("considering deconfiguration of protected\n"
+                  " package %s, to enable %s"),
+                pkg_name(pkg, pnaw_nonambig), action);
+      } else {
+        notice(_("no, %s is protected, will not deconfigure\n"
+                 " it in order to enable %s"),
+               pkg_name(pkg, pnaw_nonambig), action);
+        return 0;
+      }
+    }
+
     enqueue_deconfigure(pkg, removal);
     return 1;
   } else {
@@ -1343,11 +1361,13 @@ void check_conflict(struct dependency *dep, struct pkginfo *pkg,
       fixbyrm= dep->up;
       ensure_package_clientdata(fixbyrm);
     }
-    if (((pkg->available.essential && fixbyrm->installed.essential) ||
-         (((fixbyrm->want != PKG_WANT_INSTALL &&
-            fixbyrm->want != PKG_WANT_HOLD) ||
-           does_replace(pkg, &pkg->available, fixbyrm, &fixbyrm->installed)) &&
-          (!fixbyrm->installed.essential || in_force(FORCE_REMOVE_ESSENTIAL))))) {
+    if (((pkg->available.essential || pkg->available.is_protected) &&
+         (fixbyrm->installed.essential || fixbyrm->installed.is_protected)) ||
+        (((fixbyrm->want != PKG_WANT_INSTALL &&
+           fixbyrm->want != PKG_WANT_HOLD) ||
+          does_replace(pkg, &pkg->available, fixbyrm, &fixbyrm->installed)) &&
+         ((!fixbyrm->installed.essential || in_force(FORCE_REMOVE_ESSENTIAL)) ||
+          (!fixbyrm->installed.is_protected || in_force(FORCE_REMOVE_PROTECTED))))) {
       if (fixbyrm->clientdata->istobe != PKG_ISTOBE_NORMAL &&
           fixbyrm->clientdata->istobe != PKG_ISTOBE_DECONFIGURE)
         internerr("package %s to be fixed by removal is not to be normal "
@@ -1443,6 +1463,7 @@ void cu_cidir(int argc, void **argv) {
   char *cidirrest= (char*)argv[1];
   cidirrest[-1] = '\0';
   path_remove_tree(cidir);
+  free(cidir);
 }
 
 void cu_fileslist(int argc, void **argv) {
