@@ -212,6 +212,7 @@ static int quietmode = 0;
 static int exitnodo = 1;
 static bool background = false;
 static bool close_io = true;
+static const char *output_io;
 static bool notify_await = false;
 static int notify_timeout = 60;
 static char *notify_sockdir;
@@ -323,16 +324,15 @@ warning(const char *format, ...)
 	va_end(arglist);
 }
 
-static void DPKG_ATTR_NORET DPKG_ATTR_PRINTF(1)
-fatal(const char *format, ...)
+static void DPKG_ATTR_NORET DPKG_ATTR_VPRINTF(2)
+fatalv(int errno_fatal, const char *format, va_list args)
 {
-	va_list arglist;
-	int errno_fatal = errno;
+	va_list args_copy;
 
 	fprintf(stderr, "%s: ", progname);
-	va_start(arglist, format);
-	vfprintf(stderr, format, arglist);
-	va_end(arglist);
+	va_copy(args_copy, args);
+	vfprintf(stderr, format, args_copy);
+	va_end(args_copy);
 	if (errno_fatal)
 		fprintf(stderr, " (%s)\n", strerror(errno_fatal));
 	else
@@ -344,6 +344,24 @@ fatal(const char *format, ...)
 		exit(2);
 }
 
+static void DPKG_ATTR_NORET DPKG_ATTR_PRINTF(1)
+fatal(const char *format, ...)
+{
+	va_list args;
+
+	va_start(args, format);
+	fatalv(0, format, args);
+}
+
+static void DPKG_ATTR_NORET DPKG_ATTR_PRINTF(1)
+fatale(const char *format, ...)
+{
+	va_list args;
+
+	va_start(args, format);
+	fatalv(errno, format, args);
+}
+
 #define BUG(...) bug(__FILE__, __LINE__, __func__, __VA_ARGS__)
 
 static void DPKG_ATTR_NORET DPKG_ATTR_PRINTF(4)
@@ -371,7 +389,7 @@ xmalloc(int size)
 	ptr = malloc(size);
 	if (ptr)
 		return ptr;
-	fatal("malloc(%d) failed", size);
+	fatale("malloc(%d) failed", size);
 }
 
 static char *
@@ -382,7 +400,7 @@ xstrndup(const char *str, size_t n)
 	new_str = strndup(str, n);
 	if (new_str)
 		return new_str;
-	fatal("strndup(%s, %zu) failed", str, n);
+	fatale("strndup(%s, %zu) failed", str, n);
 }
 
 static void
@@ -391,12 +409,12 @@ timespec_gettime(struct timespec *ts)
 #if defined(_POSIX_TIMERS) && _POSIX_TIMERS > 0 && \
     defined(_POSIX_MONOTONIC_CLOCK) && _POSIX_MONOTONIC_CLOCK > 0
 	if (clock_gettime(CLOCK_MONOTONIC, ts) < 0)
-		fatal("clock_gettime failed");
+		fatale("clock_gettime failed");
 #else
 	struct timeval tv;
 
 	if (gettimeofday(&tv, NULL) != 0)
-		fatal("gettimeofday failed");
+		fatale("gettimeofday failed");
 
 	ts->tv_sec = tv.tv_sec;
 	ts->tv_nsec = tv.tv_usec * NANOSEC_IN_MICROSEC;
@@ -448,10 +466,10 @@ parse_unsigned(const char *string, int base, int *value_r)
 	long value;
 	char *endptr;
 
+	errno = 0;
 	if (!string[0])
 		return -1;
 
-	errno = 0;
 	value = strtol(string, &endptr, base);
 	if (string == endptr || *endptr != '\0' || errno != 0)
 		return -1;
@@ -486,7 +504,7 @@ detach_controlling_tty(void)
 		return;
 
 	if (ioctl(tty_fd, TIOCNOTTY, 0) != 0)
-		fatal("unable to detach controlling tty");
+		fatale("unable to detach controlling tty");
 
 	close(tty_fd);
 #endif
@@ -534,8 +552,8 @@ wait_for_child(pid_t pid)
 static void
 cleanup_socket_dir(void)
 {
-	unlink(notify_socket);
-	rmdir(notify_sockdir);
+	(void)unlink(notify_socket);
+	(void)rmdir(notify_sockdir);
 }
 
 static char *
@@ -552,18 +570,18 @@ setup_socket_name(const char *suffix)
 	}
 
 	if (asprintf(&notify_sockdir, "%s/%s.XXXXXX", basedir, suffix) < 0)
-		fatal("cannot allocate socket directory name");
+		fatale("cannot allocate socket directory name");
 
 	if (mkdtemp(notify_sockdir) == NULL)
-		fatal("cannot create socket directory %s", notify_sockdir);
+		fatale("cannot create socket directory %s", notify_sockdir);
 
 	atexit(cleanup_socket_dir);
 
 	if (chown(notify_sockdir, runas_uid, runas_gid))
-		fatal("cannot change socket directory ownership");
+		fatale("cannot change socket directory ownership");
 
 	if (asprintf(&notify_socket, "%s/notify", notify_sockdir) < 0)
-		fatal("cannot allocate socket name");
+		fatale("cannot allocate socket name");
 
 	setenv("NOTIFY_SOCKET", notify_socket, 1);
 
@@ -576,7 +594,7 @@ set_socket_passcred(int fd)
 #ifdef SO_PASSCRED
 	static const int enable = 1;
 
-	setsockopt(fd, SOL_SOCKET, SO_PASSCRED, &enable, sizeof(enable));
+	(void)setsockopt(fd, SOL_SOCKET, SO_PASSCRED, &enable, sizeof(enable));
 #endif
 }
 
@@ -590,16 +608,16 @@ create_notify_socket(void)
 	/* Create notification socket. */
 	fd = socket(AF_UNIX, SOCK_DGRAM | SOCK_NONBLOCK, 0);
 	if (fd < 0)
-		fatal("cannot create notification socket");
+		fatale("cannot create notification socket");
 
 	/* We could set SOCK_CLOEXEC instead, but then we would need to
 	 * check whether the socket call failed, try and then do this anyway,
 	 * when we have no threading problems to worry about. */
 	flags = fcntl(fd, F_GETFD);
 	if (flags < 0)
-		fatal("cannot read fd flags for notification socket");
+		fatale("cannot read fd flags for notification socket");
 	if (fcntl(fd, F_SETFD, flags | FD_CLOEXEC) < 0)
-		fatal("cannot set close-on-exec flag for notification socket");
+		fatale("cannot set close-on-exec flag for notification socket");
 
 	sockname = setup_socket_name(".s-s-d-notify");
 
@@ -611,15 +629,15 @@ create_notify_socket(void)
 
 	rc = bind(fd, &su, sizeof(su));
 	if (rc < 0)
-		fatal("cannot bind to notification socket");
+		fatale("cannot bind to notification socket");
 
 	rc = chmod(su.sun_path, 0660);
 	if (rc < 0)
-		fatal("cannot change notification socket permissions");
+		fatale("cannot change notification socket permissions");
 
 	rc = chown(su.sun_path, runas_uid, runas_gid);
 	if (rc < 0)
-		fatal("cannot change notification socket ownership");
+		fatale("cannot change notification socket ownership");
 
 	/* XXX: Verify we are talking to an expected child? Although it is not
 	 * clear whether this is feasible given the knowledge we have got. */
@@ -653,7 +671,7 @@ wait_for_notify(int fd)
 		/* Catch non-restartable errors, that is, not signals nor
 		 * kernel out of resources. */
 		if (rc < 0 && (errno != EINTR && errno != EAGAIN))
-			fatal("cannot monitor notification socket for activity");
+			fatale("cannot monitor notification socket for activity");
 
 		/* Timed-out. */
 		if (rc == 0)
@@ -678,7 +696,7 @@ wait_for_notify(int fd)
 
 			nrecv = recv(fd, buf, sizeof(buf), 0);
 			if (nrecv < 0 && (errno != EINTR && errno != EAGAIN))
-				fatal("cannot receive notification packet");
+				fatale("cannot receive notification packet");
 			if (nrecv < 0)
 				break;
 
@@ -694,7 +712,7 @@ wait_for_notify(int fd)
 					int extend_usec = 0;
 
 					if (parse_unsigned(line + 20, 10, &extend_usec) != 0)
-						fatal("cannot parse extended timeout notification %s", line);
+						fatale("cannot parse extended timeout notification %s", line);
 
 					/* Reset the current timeout. */
 					timeout.tv_sec = extend_usec / 1000L;
@@ -707,9 +725,9 @@ wait_for_notify(int fd)
 					int suberrno = 0;
 
 					if (parse_unsigned(line + 6, 10, &suberrno) != 0)
-						fatal("cannot parse errno notification %s", line);
+						fatale("cannot parse errno notification %s", line);
 					errno = suberrno;
-					fatal("program failed to initialize");
+					fatale("program failed to initialize");
 				} else if (strcmp(line, "READY=1") == 0) {
 					debug("-> Notification => ready for service.\n");
 					return;
@@ -734,19 +752,19 @@ write_pidfile(const char *filename, pid_t pid)
 		fp = fdopen(fd, "w");
 
 	if (fp == NULL)
-		fatal("unable to open pidfile '%s' for writing", filename);
+		fatale("unable to open pidfile '%s' for writing", filename);
 
 	fprintf(fp, "%d\n", pid);
 
 	if (fclose(fp))
-		fatal("unable to close pidfile '%s'", filename);
+		fatale("unable to close pidfile '%s'", filename);
 }
 
 static void
 remove_pidfile(const char *filename)
 {
 	if (unlink(filename) < 0 && errno != ENOENT)
-		fatal("cannot remove pidfile '%s'", filename);
+		fatale("cannot remove pidfile '%s'", filename);
 }
 
 static void
@@ -764,14 +782,14 @@ daemonize(void)
 	sigemptyset(&mask);
 	sigaddset(&mask, SIGCHLD);
 	if (sigprocmask(SIG_BLOCK, &mask, &oldmask) == -1)
-		fatal("cannot block SIGCHLD");
+		fatale("cannot block SIGCHLD");
 
 	if (notify_await)
 		notify_fd = create_notify_socket();
 
 	pid = fork();
 	if (pid < 0)
-		fatal("unable to do first fork");
+		fatale("unable to do first fork");
 	else if (pid) { /* First Parent. */
 		/* Wait for the second parent to exit, so that if we need to
 		 * perform any actions there, like creating a pidfile, we do
@@ -790,13 +808,17 @@ daemonize(void)
 		_exit(0);
 	}
 
+	/* Close the notification socket, even though it is close-on-exec. */
+	if (notify_await)
+		close(notify_fd);
+
 	/* Create a new session. */
 	if (setsid() < 0)
-		fatal("cannot set session ID");
+		fatale("cannot set session ID");
 
 	pid = fork();
 	if (pid < 0)
-		fatal("unable to do second fork");
+		fatale("unable to do second fork");
 	else if (pid) { /* Second parent. */
 		/* Set a default umask for dumb programs, which might get
 		 * overridden by the --umask option later on, so that we get
@@ -811,7 +833,7 @@ daemonize(void)
 	}
 
 	if (sigprocmask(SIG_SETMASK, &oldmask, NULL) == -1)
-		fatal("cannot restore signal mask");
+		fatale("cannot restore signal mask");
 
 	debug("Detaching complete...\n");
 }
@@ -887,6 +909,7 @@ usage(void)
 "      --notify-await            wait for a readiness notification\n"
 "      --notify-timeout <int>    timeout after <int> seconds of notify wait\n"
 "  -C, --no-close                do not close any file descriptor\n"
+"  -O, --output <filename>       send stdout and stderr to <filename>\n"
 "  -m, --make-pidfile            create the pidfile before starting\n"
 "      --remove-pidfile          delete the pidfile after stopping\n"
 "  -R, --retry <schedule>        check whether processes die, and retry\n"
@@ -1037,7 +1060,7 @@ parse_proc_schedule(const char *string)
 
 	if (string[policy_len] == ':' &&
 	    parse_unsigned(string + policy_len + 1, 10, &prio) != 0)
-		fatal("invalid process scheduler priority");
+		fatale("invalid process scheduler priority");
 
 	proc_sched = xmalloc(sizeof(*proc_sched));
 	proc_sched->policy_name = policy_str;
@@ -1069,7 +1092,7 @@ parse_io_schedule(const char *string)
 
 	if (string[class_len] == ':' &&
 	    parse_unsigned(string + class_len + 1, 10, &prio) != 0)
-		fatal("invalid IO scheduler priority");
+		fatale("invalid IO scheduler priority");
 
 	io_sched = xmalloc(sizeof(*io_sched));
 	io_sched->policy_name = class_str;
@@ -1101,7 +1124,7 @@ set_proc_schedule(struct res_schedule *sched)
 	param.sched_priority = sched->priority;
 
 	if (sched_setscheduler(getpid(), sched->policy, &param) == -1)
-		fatal("unable to set process scheduler");
+		fatale("unable to set process scheduler");
 #endif
 }
 
@@ -1260,6 +1283,7 @@ parse_options(int argc, char * const *argv)
 		{ "notify-await", 0, NULL, OPT_NOTIFY_AWAIT},
 		{ "notify-timeout", 1, NULL, OPT_NOTIFY_TIMEOUT},
 		{ "no-close",	  0, NULL, 'C'},
+		{ "output",	  1, NULL, 'O'},
 		{ "make-pidfile", 0, NULL, 'm'},
 		{ "remove-pidfile", 0, NULL, OPT_RM_PIDFILE},
 		{ "retry",	  1, NULL, 'R'},
@@ -1279,7 +1303,7 @@ parse_options(int argc, char * const *argv)
 
 	for (;;) {
 		c = getopt_long(argc, argv,
-		                "HKSVTa:n:op:qr:s:tu:vx:c:N:P:I:k:bCmR:g:d:",
+		                "HKSVTa:n:op:qr:s:tu:vx:c:N:P:I:k:bCO:mR:g:d:",
 		                longopts, NULL);
 		if (c == -1)
 			break;
@@ -1342,6 +1366,7 @@ parse_options(int argc, char * const *argv)
 			execname = optarg;
 			break;
 		case 'c':  /* --chuid <username>|<uid> */
+			free(changeuser);
 			/* We copy the string just in case we need the
 			 * argument later. */
 			changeuser_len = strcspn(optarg, ":");
@@ -1382,6 +1407,9 @@ parse_options(int argc, char * const *argv)
 		case 'C': /* --no-close */
 			close_io = false;
 			break;
+		case 'O': /* --outout <filename> */
+			output_io = optarg;
+			break;
 		case 'm':  /* --make-pidfile */
 			mpidfile = true;
 			break;
@@ -1431,6 +1459,9 @@ parse_options(int argc, char * const *argv)
 			badusage("umask value must be a positive number");
 	}
 
+	if (output_io != NULL && output_io[0] != '/')
+		badusage("--output file needs to be an absolute filename");
+
 	if (notify_timeout_str != NULL)
 		if (parse_unsigned(notify_timeout_str, 10, &notify_timeout) != 0)
 			badusage("invalid notify timeout value");
@@ -1469,6 +1500,11 @@ parse_options(int argc, char * const *argv)
 
 	if (!close_io && !background)
 		badusage("--no-close is only relevant with --background");
+	if (output_io && !background)
+		badusage("--output is only relevant with --background");
+
+	if (close_io && output_io == NULL)
+		output_io = "/dev/null";
 }
 
 static void
@@ -1487,7 +1523,7 @@ setup_options(void)
 			fullexecname = execname;
 
 		if (stat(fullexecname, &exec_stat))
-			fatal("unable to stat %s", fullexecname);
+			fatale("unable to stat %s", fullexecname);
 
 		if (fullexecname != execname)
 			free(fullexecname);
@@ -1498,7 +1534,7 @@ setup_options(void)
 
 		pw = getpwnam(userspec);
 		if (!pw)
-			fatal("user '%s' not found", userspec);
+			fatale("user '%s' not found", userspec);
 
 		user_id = pw->pw_uid;
 	}
@@ -1508,7 +1544,7 @@ setup_options(void)
 
 		gr = getgrnam(changegroup);
 		if (!gr)
-			fatal("group '%s' not found", changegroup);
+			fatale("group '%s' not found", changegroup);
 		changegroup = gr->gr_name;
 		runas_gid = gr->gr_gid;
 	}
@@ -1521,7 +1557,7 @@ setup_options(void)
 		else
 			pw = getpwnam(changeuser);
 		if (!pw)
-			fatal("user '%s' not found", changeuser);
+			fatale("user '%s' not found", changeuser);
 		changeuser = pw->pw_name;
 		runas_uid = pw->pw_uid;
 		if (changegroup == NULL) {
@@ -2226,7 +2262,7 @@ pid_is_running(pid_t pid)
 	else if (errno == ESRCH)
 		return false;
 	else
-		fatal("error checking pid %u status", pid);
+		fatale("error checking pid %u status", pid);
 }
 #endif
 
@@ -2267,22 +2303,25 @@ do_pidfile(const char *name)
 		 * contents cannot be trusted, because the daemon might have
 		 * been compromised.
 		 *
+		 * If the pidfile is world-writable we refuse to parse it.
+		 *
 		 * If we got /dev/null specified as the pidfile, we ignore the
 		 * checks, as this is being used to run processes no matter
 		 * what. */
-		if (match_mode == MATCH_PIDFILE &&
-		    strcmp(name, "/dev/null") != 0) {
+		if (strcmp(name, "/dev/null") != 0) {
 			struct stat st;
 			int fd = fileno(f);
 
 			if (fstat(fd, &st) < 0)
-				fatal("cannot stat pidfile %s", name);
+				fatale("cannot stat pidfile %s", name);
 
-			if ((st.st_uid != getuid() && st.st_uid != 0) ||
-			    (st.st_gid != getgid() && st.st_gid != 0))
+			if (match_mode == MATCH_PIDFILE &&
+			    ((st.st_uid != getuid() && st.st_uid != 0) ||
+			     (st.st_gid != getgid() && st.st_gid != 0)))
 				fatal("matching only on non-root pidfile %s is insecure", name);
 			if (st.st_mode & 0002)
-				fatal("matching only on world-writable pidfile %s is insecure", name);
+				fatal("matching on world-writable pidfile %s is insecure", name);
+
 		}
 
 		if (fscanf(f, "%d", &pid) == 1)
@@ -2298,7 +2337,7 @@ do_pidfile(const char *name)
 	} else if (errno == ENOENT)
 		return STATUS_DEAD;
 	else
-		fatal("unable to open pidfile %s", name);
+		fatale("unable to open pidfile %s", name);
 }
 
 #if defined(OS_Linux) || defined(OS_Solaris) || defined(OS_AIX)
@@ -2313,7 +2352,7 @@ do_procinit(void)
 
 	procdir = opendir("/proc");
 	if (!procdir)
-		fatal("unable to opendir /proc");
+		fatale("unable to opendir /proc");
 
 	foundany = 0;
 	while ((entry = readdir(procdir)) != NULL) {
@@ -2503,6 +2542,7 @@ static int
 do_start(int argc, char **argv)
 {
 	int devnull_fd = -1;
+	int output_fd = -1;
 	gid_t rgid;
 	uid_t ruid;
 
@@ -2539,6 +2579,8 @@ do_start(int argc, char **argv)
 		return 0;
 	debug("Starting %s...\n", startas);
 	*--argv = startas;
+	if (umask_value >= 0)
+		umask(umask_value);
 	if (background)
 		/* Ok, we need to detach this process. */
 		daemonize();
@@ -2546,36 +2588,39 @@ do_start(int argc, char **argv)
 		/* User wants _us_ to make the pidfile, but detach themself! */
 		write_pidfile(pidfile, getpid());
 	if (background && close_io) {
-		devnull_fd = open("/dev/null", O_RDWR);
+		devnull_fd = open("/dev/null", O_RDONLY);
 		if (devnull_fd < 0)
-			fatal("unable to open '%s'", "/dev/null");
+			fatale("unable to open '%s'", "/dev/null");
+	}
+	if (background && output_io) {
+		output_fd = open(output_io, O_CREAT | O_WRONLY, 0664);
+		if (output_fd < 0)
+			fatale("unable to open '%s'", output_io);
 	}
 	if (nicelevel) {
 		errno = 0;
 		if ((nice(nicelevel) == -1) && (errno != 0))
-			fatal("unable to alter nice level by %i", nicelevel);
+			fatale("unable to alter nice level by %i", nicelevel);
 	}
 	if (proc_sched)
 		set_proc_schedule(proc_sched);
 	if (io_sched)
 		set_io_schedule(io_sched);
-	if (umask_value >= 0)
-		umask(umask_value);
 	if (changeroot != NULL) {
 		if (chdir(changeroot) < 0)
-			fatal("unable to chdir() to %s", changeroot);
+			fatale("unable to chdir() to %s", changeroot);
 		if (chroot(changeroot) < 0)
-			fatal("unable to chroot() to %s", changeroot);
+			fatale("unable to chroot() to %s", changeroot);
 	}
 	if (chdir(changedir) < 0)
-		fatal("unable to chdir() to %s", changedir);
+		fatale("unable to chdir() to %s", changedir);
 
 	rgid = getgid();
 	ruid = getuid();
 	if (changegroup != NULL) {
 		if (rgid != (gid_t)runas_gid)
 			if (setgid(runas_gid))
-				fatal("unable to set gid to %d", runas_gid);
+				fatale("unable to set gid to %d", runas_gid);
 	}
 	if (changeuser != NULL) {
 		/* We assume that if our real user and group are the same as
@@ -2583,27 +2628,29 @@ do_start(int argc, char **argv)
 		 * will be already in place. */
 		if (rgid != (gid_t)runas_gid || ruid != (uid_t)runas_uid)
 			if (initgroups(changeuser, runas_gid))
-				fatal("unable to set initgroups() with gid %d",
+				fatale("unable to set initgroups() with gid %d",
 				      runas_gid);
 
 		if (ruid != (uid_t)runas_uid)
 			if (setuid(runas_uid))
-				fatal("unable to set uid to %s", changeuser);
+				fatale("unable to set uid to %s", changeuser);
 	}
 
+	if (background && output_fd >= 0) {
+		dup2(output_fd, 1); /* stdout */
+		dup2(output_fd, 2); /* stderr */
+	}
 	if (background && close_io) {
 		int i;
 
 		dup2(devnull_fd, 0); /* stdin */
-		dup2(devnull_fd, 1); /* stdout */
-		dup2(devnull_fd, 2); /* stderr */
 
 		 /* Now close all extra fds. */
 		for (i = get_open_fd_max() - 1; i >= 3; --i)
 			close(i);
 	}
 	execv(startas, argv);
-	fatal("unable to start %s", startas);
+	fatale("unable to start %s", startas);
 }
 
 static void
@@ -2665,7 +2712,7 @@ set_what_stop(const char *format, ...)
 	va_end(arglist);
 
 	if (rc < 0)
-		fatal("cannot allocate formatted string");
+		fatale("cannot allocate formatted string");
 }
 
 /*
@@ -2730,7 +2777,7 @@ do_stop_timeout(int timeout, int *n_killed, int *n_notkilled)
 
 		rc = pselect(0, NULL, NULL, NULL, &interval, NULL);
 		if (rc < 0 && errno != EINTR)
-			fatal("select() failed for pause");
+			fatale("select() failed for pause");
 	}
 }
 
